from typing import Any

import matplotlib
import numpy as np
import pandas as pd
import plotext as plt
import xarray as xr


def plot_terminal(
    lon: float,
    lat: float,
    dataset: xr.Dataset,
    variable: str = "v",
    operation: str = "median",
    freq: str = "m",
) -> None:
    """Plots a variable from an already located xarray dataset in the terminal"""
    plt.date_form("Y-m-d")
    sat = np.unique(dataset["satellite_img1"].values)
    sats = set([str(s[0]) for s in sat])
    g_ts = dataset.to_pandas().sort_index()

    for satellite in sats:
        ts = getattr(
            (
                g_ts.where(g_ts["satellite_img1"].str[:1] == satellite)[
                    variable
                ].groupby(pd.Grouper(freq=freq))
            ),
            operation,
        )()

        dates = [
            d.to_pydatetime().strftime("%Y-%m-%d")
            for d in pd.date_range(start=ts.index.min(), end=ts.index.max(), freq=freq)
        ]

        values = [float(v[0]) for v in ts.values]
        if satellite in ["1", "2"]:
            label = f"Sentinel {satellite}"
        else:
            label = f"Landsat {satellite}"
        plt.plot(dates, values, label=label)

        xticks, xlabels = list(dates[::6]), list(dates[::6])
        plt.xticks(xticks, xlabels)
    plt.xlabel("Date")
    # TODO: add lat, lon in plot
    plt.ylabel(variable)
    plt.show()


def _plot_by_location(
    lon: float,
    lat: float,
    ax: matplotlib.axes.Axes,
    dataset: xr.Dataset,
    variable: str = "v",
) -> None:
    dt = dataset["date_dt"].values
    dt = dt.astype(float) * 1.15741e-14

    ax.plot(
        dataset["mid_date"],
        dataset[variable],
        label=f"Lat: {lat},  Lon: {lon}",
    )

    return ax


def _plot_by_sensor(
    lon: float,
    lat: float,
    ax: matplotlib.axes.Axes,
    dataset: xr.Dataset,
    variable: str = "v",
) -> None:

    sats = np.unique(dataset["satellite_img1"].values)
    sat_plotsym_dict = {
        "1": "r+",
        "2": "bo",
        "4": "y+",
        "5": "y+",
        "7": "c+",
        "8": "g*",
        "9": "m^",
    }

    sat_label_dict = {
        "1": "Sentinel 1",
        "2": "Sentinel 2",
        "4": "Landsat 4",
        "5": "Landsat 5",
        "7": "Landsat 7",
        "8": "Landsat 8",
        "9": "Landsat 9",
    }

    ax.set_xlabel("Date")
    ax.set_ylabel("Speed (m/yr)")
    ax.set_title("ITS_LIVE Ice Flow Speed m/yr")

    dt = dataset["date_dt"].values
    dt = dt.astype(float) * 1.15741e-14

    for satellite in sats:
        ax.plot(
            dataset["mid_date"].where((dataset["satellite_img1"] == satellite)),
            dataset[variable].where((dataset["satellite_img1"] == satellite)),
            sat_plotsym_dict[satellite],
            markersize=3,
            label=sat_label_dict[satellite],
        )

    return ax


def plot_variable(
    lon: float,
    lat: float,
    ax: matplotlib.axes.Axes,
    dataset: xr.Dataset,
    variable: str = "v",
    label_by: str = "location",
) -> Any:
    """ """
    if label_by == "location":
        return _plot_by_location(lon, lat, ax, dataset, variable)
    else:
        return _plot_by_sensor(lon, lat, ax, dataset, variable)



# to get and use geojson datacube catalog
# for timing data access
from pathlib import Path
from typing import Any, Dict, List, Optional
from uuid import uuid4

import pyproj
import requests
# for datacube xarray/zarr access
import xarray as xr
from matplotlib import pyplot as plt
from rich import print as rprint
from rich.progress import track
from shapely import geometry

import itslive

from .data_viz import plot_terminal, plot_variable


# class to throw time series lookup errors
class timeseriesException(Exception):
    pass


DEFAULT_CATALOG_URL = (
    "https://its-live-data.s3.amazonaws.com/datacubes/catalog_v02.json"
)


# keep track of open cubes so that we don't re-read xarray metadata
# and dimension vectors
_open_cubes = {}
_catalog = requests.get(DEFAULT_CATALOG_URL).json()
_current_catalog_url = DEFAULT_CATALOG_URL


def _get_projected_xy_point(lon: float, lat: float, projection: str) -> geometry.Point:
    reprojection = pyproj.Transformer.from_proj(
        "epsg:4326", f"epsg:{projection}", always_xy=True
    )
    point = geometry.Point(*reprojection.transform(lon, lat))
    return point


def _merge_default_variables(variables: List[str]) -> set[str]:
    _default_variables = [
        "v",
        "v_error",
        "vx",
        "vx_error",
        "vy",
        "vy_error",
        "date_dt",
        "satellite_img1",
        "mission_img1",
    ]
    query_variables = set(_default_variables)
    query_variables.update(variables)
    return query_variables

    return []


def list_variables() -> None:
    variables = itslive.__variables__
    for v in variables:
        rprint(f"[red on black] {v} [/]")


def find(
    points: List[tuple[float, float]],
) -> List[Dict[str, Any]]:
    """Find the zarr cube information for a given geometry, if 2 values are passed
    it will use the point geometry, if 3 or more values are passed it will
    search using a polygon.

    :param first:
        Longitude value in 4326 format e.g. 28.1
    :type first: ``float``
    :param second:
        Latitude value in 4326 format e.g. -70.2
    :type second: ``float``
    """
    cubes: List = []
    if len(points) == 1:
        point = points[0]
        cubes = find_by_point(lon=point[0], lat=point[1])
    if len(points) > 1:
        cubes = find_by_polygon(points)
    return cubes


def load_catalog(
    url: str = "https://its-live-data.s3.amazonaws.com/datacubes/catalog_v02.json",
):
    """Loads a geojson catalog containing all the zarr cube urls and metadata,
    if url is not provided will load the default location on S3
    """
    try:
        _current_catalog_url = url
        _catalog = requests.get(_current_catalog_url).json()
    except Exception:
        raise Exception
    return _catalog


def find_by_bbox(
    lower_left_lon: float,
    lower_left_lat: float,
    upper_right_lon: float,
    upper_right_lat: float,
) -> List[Dict[str, Any]]:
    """
    Finds the zarr cubes that intersect with a given bounding box
    and returns a list of URLs.

    :param lower_left_lon: lower left longitude
    :param lower_left_lat: lower left longitude
    :param upper_right_lon: lower left longitude
    :param upper_right_lat: lower left longitude
    :returns: list of URLs for the matching Zarr cubes.
    """
    cubes: List = []
    box_to_find = geometry.box(
        lower_left_lon, lower_left_lat, upper_right_lon, upper_right_lat
    )
    for f in _catalog["features"]:
        polygeom = geometry.shape(f["geometry"])
        if polygeom.intersects(box_to_find):
            cubes.append(f)
    return cubes


def find_by_point(lon: float, lat: float) -> List[Dict[str, Any]]:
    """
    Finds the zarr cubes that contain a given lon, lat pair.

    :param lon: longitude
    "param lat: latitude
    :returns: geojson dictionary with matching cube
    """
    cubes: List = []
    point = geometry.Point(lon, lat)
    for f in _catalog["features"]:
        polygeom = geometry.shape(f["geometry"])
        if polygeom.contains(point):
            cubefeature = f
            projected_bbox = geometry.shape(cubefeature["properties"]["geometry_epsg"])
            projected_point = _get_projected_xy_point(
                lon, lat, cubefeature["properties"]["epsg"]
            )
            cubes.append(cubefeature)
            if not projected_bbox.contains(projected_point):
                print(
                    "Warning: bbox in projected coordinates does not contain selected point"
                )
                # TODO: implement Mark's fix to find the closest cube
            break

    return cubes


def find_by_polygon(points: List[tuple[float, float]] = []) -> List[Dict[str, Any]]:
    """
    Finds the zarr cubes that contain a given polygon.

    :param points: list of polygon points i.e. [(20.1,80.0),(21.1,81.1), ...]
    :returns: list of URLs for the matching Zarr cubes
    """
    cubes: List = []
    polygon = geometry.Polygon(points)
    for f in _catalog["features"]:
        polygeom = geometry.shape(f["geometry"])
        if polygeom.intersects(polygon):
            cubes.append(f)
    return cubes


def get_time_series(
    points: List[tuple[float, float]], variables: List[str] = ["v"]
) -> List[Dict[str, Any]]:
    """
    Returns an xarray DataArray (time series) for each variable on the list for each of the lon lat points.

    :params points: List of lon lat coordinates (i.e. some points laong the center line of a glacier)
    :params variables: list of variables included in the Dataset: v, vx, vy etc.
    :returns List[Dict]: list of tuples with coordinates and xarray Datasets for the matching Zarr cubes
    """
    velocity_ts: List = []
    for point in points:
        lon = point[0]
        lat = point[1]
        results = find_by_point(lon, lat)
        if len(results):
            cube = results[0]
            projection = cube["properties"]["epsg"]
            # for zarr store modify URL for use in boto open
            zarr_url = cube["properties"]["zarr_url"]
            cube_s3_url = zarr_url.replace("http:", "s3:").replace(
                ".s3.amazonaws.com", ""
            )
            projected_point = _get_projected_xy_point(lon, lat, projection)
            # if we have already opened this cube, don't open it again
            if len(_open_cubes) and cube_s3_url in _open_cubes.keys():
                xr_da = _open_cubes[cube_s3_url]
            else:
                xr_da = xr.open_dataset(
                    cube_s3_url, engine="zarr", storage_options={"anon": True}
                )
                _open_cubes[cube_s3_url] = xr_da
            time_series = xr_da[variables].sel(
                x=projected_point.x, y=projected_point.y, method="nearest"
            )
            if time_series is not None and isinstance(time_series, xr.Dataset):
                velocity_ts.append(
                    {"coordinates": (lon, lat), "time_series": time_series}
                )

    return velocity_ts


def export_csv(
    points: List[tuple[float, float]],
    variables: List[str] = ["v"],
    outdir: Optional[str] = None,
) -> None:
    """Exports a list of ITS_LIVE glacier velocity variables to csv files

    :params points: List of lon lat coordinates (i.e. some points laong the center line of a glacier)
    :params variables: list of variables included in the Dataset: v, vx, vy etc.
    :params outdir: directory to place the CSV files
    """

    query_variables = _merge_default_variables(variables)

    outdir = f"./itslive-{uuid4()}" if outdir is None else outdir

    Path(outdir).mkdir(parents=True, exist_ok=True)

    for point in track(
        points,
        description=f"Processing {len(points)} coordinates...",
        total=len(points),
    ):

        lon = round(point[0], 4)
        lat = round(point[1], 4)
        result_series = get_time_series([(lon, lat)], query_variables)
        if len(result_series):
            series = result_series[0]["time_series"]

            df = series.to_dataframe()
            df["x"] = lon
            df["y"] = lat
            df = df.rename(
                columns={
                    "x": "lon",
                    "y": "lat",
                    "satellite_img1": "satellite",
                    "mission_img1": "mission",
                    "v": "v [m/yr]",
                    "v_error": "v_error [m/yr]",
                    "vx": "vx [m/yr]",
                    "vx_error": "vx_error [m/yr]",
                    "vy": "vy [m/yr]",
                    "vy_error": "vy_error [m/yr]",
                }
            )
            df["epsg"] = series.attrs["projection"]
            df["date_dt [days]"] = df["date_dt"].dt.days
            ts = df.dropna()
            file_name = f"LON{lon}--LAT{lat}.csv"
            ts.to_csv(
                f"{outdir}/{file_name}",
                columns=[
                    "lon",
                    "lat",
                    "v [m/yr]",
                    "v_error [m/yr]",
                    "vx [m/yr]",
                    "vx_error [m/yr]",
                    "vy [m/yr]",
                    "vy_error [m/yr]",
                    "date_dt [days]",
                    "mission",
                    "satellite",
                    "epsg",
                ],
            )
        else:
            rprint(f"[red on black]No data found at[/] lon: {lon}, lat: {lat}")


def export_netcdf(
    points: List[tuple[float, float]],
    variables: List[str] = ["v"],
    outdir: Optional[str] = None,
) -> None:
    """Exports a list of ITS_LIVE glacier velocity variables to netcdf files

    :params points: List of lon lat coordinates (i.e. some points laong the center line of a glacier)
    :params variables: list of variables included in the Dataset: v, vx, vy etc.
    :params outdir: directory to place the CSV files
    """

    query_variables = _merge_default_variables(variables)

    outdir = f"./itslive-{uuid4()}" if outdir is None else outdir
    Path(outdir).mkdir(parents=True, exist_ok=True)

    for point in track(
        points,
        description=f"Processing {len(points)} coordinates...",
        total=len(points),
    ):

        lon = round(point[0], 4)
        lat = round(point[1], 4)

        file_name = f"LON{lon}--LAT{lat}"
        result_series = get_time_series([(lon, lat)], query_variables)
        if len(result_series):
            series = result_series[0]["time_series"]
            series.to_netcdf(f"{outdir}/{file_name}.nc")
        else:
            rprint(f"[red on black]No data found at[/] lon:{lon}, lat: {lat}")


def export_stdout(
    points: List[tuple[float, float]],
    variables: List[str] = ["v"],
) -> None:
    """Exports a list of ITS_LIVE glacier velocity variables to stdout

    :params points: List of lon lat coordinates (i.e. some points laong the center line of a glacier)
    :params variables: list of variables included in the Dataset: v, vx, vy etc.
    """

    query_variables = _merge_default_variables(variables)

    for point in track(
        points,
        description=f"Processing {len(points)} coordinates...",
        total=len(points),
    ):

        lon = round(point[0], 4)
        lat = round(point[1], 4)

        result_series = get_time_series([(lon, lat)], query_variables)
        if len(result_series):
            series = result_series[0]["time_series"]
            df = series.to_dataframe()
            df["x"] = lon
            df["y"] = lat
            df = df.rename(
                columns={
                    "x": "lon",
                    "y": "lat",
                    "satellite_img1": "satellite",
                    "mission_img1": "mission",
                    "v": "v [m/yr]",
                    "v_error": "v_error [m/yr]",
                    "vx": "vx [m/yr]",
                    "vx_error": "vx_error [m/yr]",
                    "vy": "vy [m/yr]",
                    "vy_error": "vy_error [m/yr]",
                }
            )
            df["epsg"] = series.attrs["projection"]
            df["date_dt [days]"] = df["date_dt"].dt.days
            ts = df.dropna()
            outstr = ts.to_markdown()
            rprint(outstr)
        else:
            rprint(f"[red on black] No data found at [/] lon: {lon}, lat: {lat}")


def plot_time_series(
    points: List[tuple[float, float]],
    variable: str = "v",
    label_by: str = "location",
) -> Any:
    """Plots velocity time series for a list of lon, lat locations. Color coding
    can be by location or satellite, satellite is better suited for single points.

    :params points: List of lon lat coordinates (i.e. some points laong the center line of a glacier)
    :params variables: list of variables included in the Dataset: v, vx, vy etc.
    """
    fig, ax = plt.subplots(1, 1)
    for point in track(
        points,
        description=f"Processing {len(points)} coordinates...",
        total=len(points),
    ):
        lon = round(point[0], 4)
        lat = round(point[1], 4)

        query_variables = _merge_default_variables(variable)

        series = get_time_series([(lon, lat)], variables=query_variables)
        if series is not None and len(series) > 0:
            ts = series[0]["time_series"]
            plot_variable(lon, lat, ax, ts, variable, label_by)

    return ax


def _plot_time_series_terminal(
    points: List[tuple[float, float]],
    variable: List[str] = ["v"],
    operation: str = "median",
    freq: str = "m",
):
    """
    A bit of an easter egg. Plots velocity time series directly on the terminal.
    only used by the CLI via the itslive-plot command.
    """
    for point in track(
        points,
        description=f"Processing {len(points)} coordinates...",
        total=len(points),
    ):
        lon = round(point[0], 4)
        lat = round(point[1], 4)

        query_variables = _merge_default_variables(variable)

        series = get_time_series([(lon, lat)], variables=query_variables)
        if series is not None and len(series) > 0:
            ts = series[0]["time_series"]

            plot_terminal(lon, lat, ts, variable, operation, freq)

            max_variable = (
                ts[variable]
                .where(ts[variable] == ts[variable].max(), drop=True)
                .squeeze()
            )
            min_variable = (
                ts[variable]
                .where(ts[variable] == ts[variable].min(), drop=True)
                .squeeze()
            )
            max_value = max_variable[variable[0]].values
            min_value = min_variable[variable[0]].values

            rprint(f"Max {variable} on {max_variable['mid_date'].values}: {max_value}")
            rprint(f"Min {variable} on {min_variable['mid_date'].values}: {min_value}")
            rprint(f"Cube URL: {max_variable.attrs['url']}")
    return None



import itslive
import pandas as pd
import rich_click as click
from rich import print as rprint

# Use Rich markup
click.rich_click.USE_RICH_MARKUP = True


class Mutex(click.Option):
    # Taken from https://bit.ly/3hWeXVj wonder if this is already in the library
    def __init__(self, *args, **kwargs):
        self.not_required_if: list = kwargs.pop("not_required_if")

        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs["help"] = (
            f"{kwargs.get('help', '')} Option is mutually exclusive with "
            f"{', '.join(self.not_required_if).strip()}."
        )
        super(Mutex, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt: bool = self.name in opts
        for mutex_opt in self.not_required_if:
            if mutex_opt in opts:
                if current_opt:
                    error_msg = (
                        f"Illegal usage: {self.name}"
                        f" is mutually exclusive with {str(mutex_opt)}"
                    )
                    raise click.UsageError(error_msg)
                else:
                    self.prompt = None
        return super(Mutex, self).handle_parse_result(ctx, opts, args)


def validate_latitude(lctx, param, lat):
    if lat:
        if lat > 90.0 or lat < -90.0:
            error_msg = (
                f"Not a valid longitude value: {lat}, must be between -90 and 90"
            )
            raise click.BadParameter(error_msg)
        return lat


def validate_longitude(ctx, param, lon):
    if lon:
        if lon > 180.0 or lon < -180.0:
            error_msg = (
                f"Not a valid longitude value: {lon}, must be between -180 and 180"
            )
            raise click.BadParameter(error_msg)
        return lon


def validate_csv(ctx, param, value):
    if value:
        try:
            df = pd.read_csv(value, usecols=[0, 1], names=["lon", "lat"])
            assert df.lat.dtype == "float"
            assert df.lon.dtype == "float"
            assert df.lon.between(-180.0, 180.0).any()
            assert df.lat.between(-90.0, 90.0).any()
            return df
        except Exception:
            raise click.BadParameter("Not a valid CSV file, the format is lon,lat")


def export_time_series(points, variables, format, outdir):
    if format == "csv":
        itslive.cubes.export_csv(points, variables, outdir)
    elif format == "netcdf":
        itslive.cubes.export_netcdf(points, variables, outdir)
    else:
        itslive.cubes.export_stdout(points, variables)
    return None


def plot_time_series(points, variable, label_by, outdir, stdout):
    if stdout is not None:
        # experimental
        itslive.cubes._plot_time_series_terminal(points, variable)
    else:
        pass
        # TODO: save plot on outdir
        # plot = itslive.cubes.plot_time_series(points, variable, label_by)
        # plot.save()
    return None


@click.command()
@click.option(
    "--itslive-catalog",
    required=False,
    default="https://its-live-data.s3.amazonaws.com/datacubes/catalog_v02.json",
    help="GeoJSON catalog with the ITS_LIVE cube metadata",
)
@click.option(
    "--input-coordinates",
    required=False,
    cls=Mutex,
    not_required_if=["lat", "lon"],
    type=click.Path(),
    callback=validate_csv,
    help="[magenta bold]Input csv file[/]. [dim] \[format: comma separated lon,lat coordinates][/]",
)
@click.option(
    "--lat",
    cls=Mutex,
    not_required_if=["input_coordinates"],
    type=float,
    callback=validate_latitude,
    help="Latitude, e.g. 70.1",
)
@click.option(
    "--lon",
    cls=Mutex,
    not_required_if=["input_coordinates"],
    type=float,
    callback=validate_longitude,
    help="Longitude, e.g. -120.4",
)
@click.option(
    "--variables",
    type=click.Choice(["v", "v_error", "vy", "vx"], case_sensitive=False),
    multiple=True,
    default=["v"],
    help=(
        "v: velocity"
        "v_error: error in v"
        "vx: Velocity x component"
        "vy: Velocity y component"
        "for a full list visit"
    ),
)
@click.option(
    "--outdir",
    type=str,
    help="output directory",
)
@click.option(
    "--format",
    type=click.Choice(["csv", "netcdf", "stdout"]),
    help="export to fortmat",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Verbose output",
)
def export(
    itslive_catalog, input_coordinates, lat, lon, variables, outdir, format, debug
):
    """
    ITS_LIVE Global Glacier Veolocity

    [i]You can try using --help at the top level and also for
    specific group subcommands.[/]
    """

    points = []
    itslive.cubes.load_catalog(itslive_catalog)
    if debug:
        rprint("Debug mode is [red]on[/]")
        rprint(f"Using: {itslive.cubes._current_catalog_url}")
    if input_coordinates is not None:
        # rprint(f"input file head: {input.head}")
        for index, row in input_coordinates.iterrows():
            points.append((row["lon"], row["lat"]))
    else:
        if lat and lon:
            points.append((lon, lat))

    if len(points) and format is not None:
        export_time_series(points, variables, format, outdir)
    else:
        rprint(" At least one set of coordinates are needed, --help")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['itslive', 'itslive.cli']

package_data = \
{'': ['*']}

install_requires = \
['Shapely>=1.8',
 'matplotlib>=3.5',
 'pandas==1.5.1',
 'plotext>=0',
 'pyproj>=3.3',
 'requests>=2.28.1,<3.0.0',
 'rich-click>=1.5',
 's3fs>=2022.3',
 'tabulate>=0.9.0',
 'xarray>=2022.3',
 'zarr>=2.11']

entry_points = \
{'console_scripts': ['itslive-export = itslive.cli.export:export',
                     'itslive-plot = itslive.cli.plot:plot']}

setup_kwargs = {
    'name': 'itslive',
    'version': '0.1.6',
    'description': 'Python client for ITSLIVE gralcier velocity data',
    'long_description': 'None',
    'author': 'Luis Lopez',
    'author_email': 'luis.lopez@nsidc.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nasa-jpl/itslive-vortex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import pkg_resources  # type: ignore

from itslive import data_cube as cubes

__all__ = ["cubes"]

__variables__ = [
    "acquisition_date_img1",
    "acquisition_date_img2",
    "autoRIFT_software_version",
    "chip_size_height",
    "chip_size_width",
    "date_center",
    "date_dt",
    "granule_url",
    "interp_mask",
    "mapping",
    "mid_date",
    "mission_img1",
    "mission_img2",
    "roi_valid_percentage",
    "satellite_img1",
    "satellite_img2",
    "sensor_img1",
    "sensor_img2",
    "stable_count_mask",
    "stable_count_slow",
    "stable_shift_flag",
    "v",
    "v_error",
    "va",
    "va_error",
    "va_error_mask",
    "va_error_modeled",
    "va_error_slow",
    "va_stable_shift",
    "va_stable_shift_mask",
    "va_stable_shift_slow",
    "vr",
    "vr_error",
    "vr_error_mask",
    "vr_error_modeled",
    "vr_error_slow",
    "vr_stable_shift",
    "vr_stable_shift_mask",
    "vr_stable_shift_slow",
    "vx",
    "vx_error",
    "vx_error_mask",
    "vx_error_modeled",
    "vx_error_slow",
    "vx_stable_shift",
    "vx_stable_shift_mask",
    "vx_stable_shift_slow",
    "vy",
    "vy_error",
    "vy_error_mask",
    "vy_error_modeled",
    "vy_error_slow",
    "vy_stable_shift",
    "vy_stable_shift_mask",
    "vy_stable_shift_slow",
    "x",
    "y",
]

# this comes from the installed version not the editable source
__version__ = pkg_resources.get_distribution("itslive").version
